import time
import sys
from dataclasses import dataclass

from typing import List, Dict

import pytest
from from_dict import from_dict

# ...................................................................
# These allow the test to disable caching so we can verify the 
# performance gains.

import from_dict._from_dict as _fd
hints_cached_func  = _fd.get_constructor_type_hints
hints_wrapped_func  = _fd.get_constructor_type_hints.__wrapped__

fwd_ref_cached_func  = _fd._resolve_str_forward_ref
fwd_ref_wrapped_func = _fd._resolve_str_forward_ref.__wrapped__

def cache_disable():
    global hints_wrapped_func
    global fwd_ref_wrapped_func
    _fd.get_constructor_type_hints = hints_wrapped_func
    _fd._resolve_str_forward_ref = fwd_ref_wrapped_func

def cache_is_enabled():
    return _fd.get_constructor_type_hints == hints_cached_func

def cache_enable():
    global hints_cached_func
    global fwd_ref_cached_func
    _fd.get_constructor_type_hints = hints_cached_func
    _fd._resolve_str_forward_ref = fwd_ref_cached_func
    hints_cached_func.cache_clear()
    fwd_ref_cached_func.cache_clear()

# ...................................................................

@dataclass(frozen=True)
class MainData:
    name: str
    id: int
    date: str
    inner: List['InnerData']
    properties: Dict[str, str]


@dataclass(frozen=True)
class InnerData:
    name: str
    result: bool
    tags: List[str]

def test_cache_speed_improvement():
    test_data =  {
        'name': 'root', 
        'id': 12324, 
        'date': '2022-04-06T010203', 
        'inner': [
            {'name': 'Run-1', 'result': True, 'tags': ['rc']}, 
            {'name': 'Run-2', 'result': False, 'tags': ['nighty']},
            {'name': 'Run-3', 'result': True, 'tags': ['nighty']}, 
            {'name': 'Run-4', 'result': True, 'tags': ['official']}
        ], 
        'properties': {
            'prop-1': 'value', 
            'prop-2': 'value',
            'prop-3': 'value', 
            'prop-4': 'value', 
            'prop-5': 'value'
        }
    }
    cache_disable()
    assert not cache_is_enabled()
    disabled_time = _get_time_for_x_calls(test_data)
    
    cache_enable()
    assert cache_is_enabled()
    enabled_time = _get_time_for_x_calls(test_data)
    print(f"Improvement = {disabled_time / enabled_time:0.2f}")
    assert (disabled_time / enabled_time) > 2.0

    cache_disable()
    assert not cache_is_enabled()
    disabled_time = _get_time_for_x_calls(test_data)
    cache_enable()
    assert disabled_time > (enabled_time * 2)

def _get_time_for_x_calls(test_data: dict):
    start_time = time.time()
    LOOPS = 1_000
    for _ in range(LOOPS):
        from_dict(MainData, test_data)
    elapsed = time.time() - start_time
    print(f"Took {elapsed:0.3f} seconds for {LOOPS} calls")
    return elapsed


from dataclasses import dataclass
from typing import List, Optional

from from_dict import from_dict


@dataclass(frozen=True)
class Preference:
    name: str
    score: int


@dataclass(frozen=True)
class Customer:
    name: str
    nick_name: Optional[str]
    preferences: List[Preference]


input_customer_data = {
    "name": "Christopher Lee",
    "nick_name": None,
    "preferences": [
        {"name": "The Hobbit", "score": 37},
        {"name": "Count Dooku", "score": 2},
        {"name": "Saruman", "score": 99}
    ],
    "friend": "Mellon"
}


def test_readme():
    customer = from_dict(Customer, input_customer_data)
    # Structured data is available as attributes since attr.s exposes them like that
    assert customer.name == "Christopher Lee"
    # Nested structures are also constructed. List[sub_strucutre] and Dict[key, sub_structure] are supported
    assert customer.preferences[0].name == "The Hobbit"
    # Data not defined in the strucutre is inserted into the __dict__ if possible
    assert customer.__dict__["friend"] == "Mellon"



from dataclasses import dataclass
from typing import Optional, Union

import pytest
from from_dict import FromDictTypeError, from_dict


@dataclass(frozen=True)
class SimpleNode:
    name: str
    value: str


@dataclass(frozen=True)
class SimpleNode2:
    name: str
    reference: str

@dataclass
class SimpleNode3:
    name: str
    field: str = "Default-value"

@dataclass(frozen=True)
class NodeWithOptional:
    node1: SimpleNode
    node2: Optional[SimpleNode]
    node3: Optional[SimpleNode] = None


@dataclass(frozen=True)
class NodeWithUnion:
    node: Union[SimpleNode, SimpleNode2]

@dataclass(frozen=True)
class NodeWithUnionWithBuiltInType:
    node: Union[SimpleNode, str]

def test_optional():
    data = {
        "node1": {"name": "n1", "value": "v1"},
        "node2": {"name": "n2", "value": "v2"},
        "node3": {"name": "n3", "value": "v3"},
     }
    node = from_dict(NodeWithOptional, data, fd_check_types=True)
    assert node.node1.name == "n1"
    assert node.node1.value == "v1"
    assert isinstance(node.node2, SimpleNode)
    assert node.node2.name == "n2"
    assert node.node2.value == "v2"
    assert isinstance(node.node3, SimpleNode)
    assert node.node3.name == "n3"
    assert node.node3.value == "v3"
    
    data = {
        "node1": {"name": "n1", "value": "v1"},
        "node2": {"name": "n2", "value": "v2"}
     }
    node = from_dict(NodeWithOptional, data, fd_check_types=True)
    assert node.node1.name == "n1"
    assert node.node1.value == "v1"
    assert isinstance(node.node2, SimpleNode)
    assert node.node2.name == "n2"
    assert node.node2.value == "v2"
    assert node.node3 == None
    
    data = {
        "node1": {"name": "n1", "value": "v1"},
        "node2": None
     }
    node = from_dict(NodeWithOptional, data, fd_check_types=True)
    assert node.node1.name == "n1"
    assert node.node1.value == "v1"
    assert node.node2 == None
    assert node.node3 == None
    data = {
        "node1": {"name": "n1", "value": "v1"},
        "node2": None,
        "node3": None
     }
    node = from_dict(NodeWithOptional, data, fd_check_types=True)
    assert node.node1.name == "n1"
    assert node.node1.value == "v1"
    assert node.node2 == None
    assert node.node3 == None
    
    data = {"node1": {"name": "n1", "value": "v1"}}
    with pytest.raises(TypeError) as e:
        node = from_dict(NodeWithOptional, data, fd_check_types=True)
    assert "missing 1 required positional argument: 'node2'" in str(e.value)


def test_union():
    data = {  "node": {"name": "n1", "value": "v1"} }
    node = from_dict(NodeWithUnion, data, fd_check_types=True)
    assert isinstance(node.node, SimpleNode)
    assert node.node.name == "n1"
    assert node.node.value == "v1"

    data = {  "node": {"name": "n1", "reference": "node-X"} }
    node = from_dict(NodeWithUnion, data, fd_check_types=True)
    assert isinstance(node.node, SimpleNode2)
    assert node.node.name == "n1"
    assert node.node.reference == "node-X"

    data = {  "node": {"name": "n1", "no_match": "node-X"} }
    node = from_dict(NodeWithUnion, data)
    assert isinstance(node.node, dict)
    assert node.node["name"] == "n1"
    assert node.node["no_match"] == "node-X"
    
    data = {  "node": {"name": "n1", "no_match": "node-X"} }
    with pytest.raises(TypeError) as e:
        node = from_dict(NodeWithUnion, data, fd_check_types=True)
    assert str(e.value) == ('For "node", expected typing.Union[test_generics_with_dataclasses.SimpleNode, '
                            'test_generics_with_dataclasses.SimpleNode2] but found <class \'dict\'>')

def test_union_2():
    @dataclass(frozen=True)
    class NodeWithUnion2:
        node: Union[SimpleNode, SimpleNode2, SimpleNode3]
    
    data = {  "node": {"name": "n1", "value": "v1"} }
    node = from_dict(NodeWithUnion2, data, fd_check_types=True)
    assert isinstance(node.node, SimpleNode)
    assert node.node.name == "n1"
    assert node.node.value == "v1"

    data = {  "node": {"name": "n1", "reference": "node-X"} }
    node = from_dict(NodeWithUnion2, data, fd_check_types=True)
    assert isinstance(node.node, SimpleNode2)
    assert node.node.name == "n1"
    assert node.node.reference == "node-X"

    data = {  "node": {"name": "n1", "field": "value"} }
    node = from_dict(NodeWithUnion2, data, fd_check_types=True)
    assert isinstance(node.node, SimpleNode3)
    assert node.node.name == "n1"
    assert node.node.field == "value"

    data = {  "node": {"name": "n1"} }
    node = from_dict(NodeWithUnion2, data, fd_check_types=True)
    assert isinstance(node.node, SimpleNode3)
    assert node.node.name == "n1"
    assert node.node.field == "Default-value"

    data = {  "node": {"name": "n1", "no_match": "node-X"} }
    node = from_dict(NodeWithUnion2, data)
    assert isinstance(node.node, dict)
    assert node.node["name"] == "n1"
    assert node.node["no_match"] == "node-X"

    data = {  "node": {"name": "n1", "no_match": "node-X"} }
    with pytest.raises(TypeError) as e:
        node = from_dict(NodeWithUnion2, data, fd_check_types=True)
    assert str(e.value) == ('For "node", expected typing.Union[test_generics_with_dataclasses.SimpleNode, '
                            'test_generics_with_dataclasses.SimpleNode2, '
                            'test_generics_with_dataclasses.SimpleNode3] but found <class \'dict\'>')

def test_union_with_builtin_type():
    data = {  "node": {"name": "n1", "value": "v1"} }
    node = from_dict(NodeWithUnionWithBuiltInType, data, fd_check_types=True)
    assert isinstance(node.node, SimpleNode)
    assert node.node.name == "n1"
    assert node.node.value == "v1"

    data = {  "node": "Hello" }
    node = from_dict(NodeWithUnionWithBuiltInType, data, fd_check_types=True)
    assert isinstance(node.node, str)
    assert node.node == "Hello"

    data = {  "node": {"name": "n1", "no_match": "node-X"} }
    node = from_dict(NodeWithUnionWithBuiltInType, data)
    assert isinstance(node.node, dict)
    assert node.node["name"] == "n1"  # type: ignore
    assert node.node["no_match"] == "node-X"  # type: ignore
    
    data = {  "node": {"name": "n1", "no_match": "node-X"} }
    with pytest.raises(FromDictTypeError) as e:
        from_dict(NodeWithUnionWithBuiltInType, data, fd_check_types=True)
    assert str(e.value) == 'For "node", expected typing.Union[test_generics_with_dataclasses.SimpleNode, str] but found <class \'dict\'>'

    data = {  "node": 123 }
    with pytest.raises(FromDictTypeError) as e:
        node = from_dict(NodeWithUnionWithBuiltInType, data, fd_check_types=True)
    assert str(e.value) == 'For "node", expected typing.Union[test_generics_with_dataclasses.SimpleNode, str] but found <class \'int\'>'



import os
import datetime
from dataclasses import dataclass

import from_dict._from_dict as _fd

# . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

def new_empty_namespaces():
    return _fd.NamespaceTypes(None, None)

def test_empty():
    obj1 = new_empty_namespaces()
    obj2 = new_empty_namespaces()
    assert (obj1 == obj2)

# . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

def new_simple_namespaces():
    return _fd.NamespaceTypes(globals(), locals())

def test_simple():
    obj1 = new_simple_namespaces()
    obj2 = new_simple_namespaces()
    assert (obj1 == obj2)

# . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

def new_with_changing_locals():
    r = os.urandom(2)
    d = datetime.datetime.now()
    return _fd.NamespaceTypes(globals(), locals())


def test_with_changing_locals():
    obj1 = new_with_changing_locals()
    obj2 = new_with_changing_locals()
    assert (obj1 == obj2)

# . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

def new_with_global_in_locals():
    g = globals()
    return _fd.NamespaceTypes(g, locals())

def test_with_global_in_locals():
    obj1 = new_with_global_in_locals()
    obj2 = new_with_global_in_locals()
    assert (obj1 == obj2)

# . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

def new_with_local_in_locals():
    ''' local self reference '''
    g = globals()
    l = locals()
    return _fd.NamespaceTypes(g, locals())

def test_with_local_in_locals():
    obj1 = new_with_local_in_locals()
    obj2 = new_with_local_in_locals()
    assert (obj1 == obj2)

# . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

def new_with_nested_dataclass():
    @dataclass 
    class NestedDC:
        name: str
    return _fd.NamespaceTypes(globals(), locals())

def test_with_nested_dataclass():
    ''' NOTE: This test expects that the namespaces are NOT equal 
    '''
    obj1 = new_with_nested_dataclass()
    obj2 = new_with_nested_dataclass()
    assert (obj1 != obj2)


# . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

@dataclass
class DummyClass1:
    name: str
    value: int

def new_with_nested_global_dataclass():
    dc = DummyClass1
    return _fd.NamespaceTypes(globals(), locals())

def test_with_nested_global_dataclass():
    obj1 = new_with_nested_global_dataclass()
    obj2 = new_with_nested_global_dataclass()
    assert isinstance(obj2.local_types, dict)
    assert obj2.local_types["dc"]  == DummyClass1
    assert (obj1 == obj2)

# . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

def new_with_nested_module_dataclass():
    from json import JSONDecoder
    return _fd.NamespaceTypes(globals(), locals())

def test_with_nested_module_dataclass():
    obj1 = new_with_nested_module_dataclass()
    obj2 = new_with_nested_module_dataclass()
    assert isinstance(obj2.local_types, dict)
    assert obj2.local_types["JSONDecoder"].__name__  == "JSONDecoder"
    assert (obj1 == obj2)

# . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

@dataclass
class DummyClass2:
    name: str
    value: int

@dataclass
class DummyClass3:
    name: str
    value: int

def new_with_copy_of_namespaces():
    return _fd.NamespaceTypes(globals().copy(), locals().copy())

def test_with_copy_of_namespaces():
    obj1 = new_with_copy_of_namespaces()
    obj2 = new_with_copy_of_namespaces()
    assert (obj1 == obj2)

    globals()["DummyClass2"] = DummyClass3
    obj2 = new_with_copy_of_namespaces()
    assert (obj1 != obj2)



from dataclasses import dataclass
from typing import Dict, List, NamedTuple, Optional, Type, Union, TypeVar, Generic

import attr
import pytest
from from_dict import FromDictTypeError, FromDictUnknownArgsError, from_dict


@dataclass
class Structures:
    outer_structure: Type
    inner_structure: Type
    does_type_validation: bool
    has_dict: bool


# Attr style structures
@attr.s(auto_attribs=True)
class SubTestDictAttr:
    foo: int
    bar: str


@attr.s(auto_attribs=True)
class MainTestDictAttr:
    foo: int
    baz: SubTestDictAttr


@attr.s(auto_attribs=True)
class SubTestDictAttrWithValidators:
    foo: int = attr.ib(validator=attr.validators.instance_of(int))
    bar: str = attr.ib(validator=attr.validators.instance_of(str))


@attr.s(auto_attribs=True)
class MainTestDictAttrWithValidators:
    foo: int = attr.ib(validator=attr.validators.instance_of(int))
    baz: SubTestDictAttrWithValidators = attr.ib(validator=attr.validators.instance_of(SubTestDictAttrWithValidators))


# Dataclass structures
@dataclass(frozen=True)
class SubTestDictDataclass:
    foo: int
    bar: str


@dataclass(frozen=True)
class MainTestDictDataclass:
    foo: int
    baz: SubTestDictDataclass


# NamedTuple structures
class SubTestDictNamedTuple(NamedTuple):
    foo: int
    bar: str


class MainTestDictNamedTuple(NamedTuple):
    foo: int
    baz: SubTestDictNamedTuple


# Dataclass structures
@dataclass(frozen=True)
class MainTestDictDataclassFrwRef:
    foo: int
    baz: 'SubTestDictDataclassFrwRef'


@dataclass(frozen=True)
class SubTestDictDataclassFrwRef:
    foo: int
    bar: str

@pytest.fixture(params=[
    pytest.param(Structures(MainTestDictAttr, SubTestDictAttr, False, True), id="attr"),
    pytest.param(Structures(MainTestDictAttr, SubTestDictAttr, True, True), id="attr-with-validators"),
    pytest.param(Structures(MainTestDictDataclass, SubTestDictDataclass, False, True), id="dataclass"),
    pytest.param(Structures(MainTestDictDataclassFrwRef, SubTestDictDataclassFrwRef, False, True), id="dataclass-forward-ref"),
    pytest.param(Structures(MainTestDictNamedTuple, SubTestDictNamedTuple, False, False), id="named-tuple"),
])
def structures(request):
    yield request.param


def test_packing(structures: Structures):
    input_dict = {
        "foo": 22,
        "baz": {
            "foo": 42,
            "bar": "Works :)",
        }
    }

    main_object = from_dict(structures.outer_structure, input_dict)

    assert main_object.baz.bar == "Works :)"
    assert isinstance(main_object, structures.outer_structure)
    assert isinstance(main_object.baz, structures.inner_structure)


def test_keyword_style(structures: Structures):
    m = from_dict(structures.outer_structure, foo=22, baz=structures.inner_structure(foo=42, bar="Works :)"))
    assert m.foo == 22
    assert m.baz.foo == 42
    assert m.baz.bar == "Works :)"


def test_keyword_style_overwrites_positional(structures: Structures):
    assert from_dict(structures.inner_structure, {"foo": 42, "bar": "Works :)"}, foo=0).foo == 0


def test_additional_keys_are_allowed(structures: Structures):
    my_obj = from_dict(structures.inner_structure, foo=22, bar="Works", additional=[1, 2, 3])
    assert my_obj.foo == 22
    assert my_obj.bar == "Works"

    if structures.has_dict:
        assert my_obj.additional == [1, 2, 3]


def test_missing_key(structures: Structures):
    with pytest.raises(TypeError):
        from_dict(structures.inner_structure, {"foo": 22})


def test_invalid_type_inherent(structures: Structures):
    if not structures.does_type_validation:
        pytest.skip("Structure does not support inherent runtime type checking")

    with pytest.raises(TypeError) as e:
        from_dict(structures.inner_structure, {"foo": "wrong", "bar": "right"}, fd_check_types=True)


def test_invalid_type_from_dict(structures: Structures):
    with pytest.raises(FromDictTypeError) as e:
        from_dict(structures.inner_structure, {"foo": "wrong", "bar": "right"}, fd_check_types=True)

    assert str(e.value) == "For \"foo\", expected <class 'int'> but found <class 'str'>"


def test_missing_key_discovered_in_subdict_inherent(structures: Structures):
    if not structures.does_type_validation:
        pytest.skip("Structure does not support inherent runtime type checking")

    with pytest.raises(TypeError):
        from_dict(structures.outer_structure, {"foo": 22, "baz": {"foo": 42}})


def test_invalid_type_discovered_in_subdict_from_dict(structures: Structures):
    with pytest.raises(FromDictTypeError) as e:
        from_dict(structures.outer_structure, {"foo": 22, "baz": {"foo": 42, "bar": ["wrong type"]}},
                  fd_check_types=True)

    assert str(e.value) == "For \"baz.bar\", expected <class 'str'> but found <class 'list'>"


def test_invalid_type_discovered_in_subdict(structures: Structures):
    with pytest.raises(FromDictTypeError) as e:
        from_dict(structures.outer_structure, {"foo": 22, "baz": {"foo": 42, "bar": ["wrong type"]}},
                  fd_check_types=True)

    assert str(e.value) == "For \"baz.bar\", expected <class 'str'> but found <class 'list'>"


def test_invalid_list_element_type():
    @dataclass(frozen=True)
    class TstClass:
        a: int
        b: Optional[str]
        c: List[int]
        d: Dict[str, int]

    with pytest.raises(FromDictTypeError) as e:
        opt = from_dict(TstClass, a=11, b=None, c=[1, 2, 3, "bad"], d={"a":1, "b": 2}, fd_check_types=True)

    assert str(e.value) == "For \"c[3]\", expected <class 'int'> but found <class 'str'>"
    assert not e.value.__suppress_context__


def test_invalid_dict_element_type():
    @dataclass(frozen=True)
    class TstClass:
        a: int
        b: Optional[str]
        c: List[int]
        d: Dict[str, int]

    with pytest.raises(FromDictTypeError) as e:
        opt = from_dict(TstClass, a=11, b=None, c=[1, 2, 3, 4], d={"a":1, "b": 2, "C": "bad"}, fd_check_types=True)

    assert str(e.value) == "For \"d['C']\", expected <class 'int'> but found <class 'str'>"
    assert not e.value.__suppress_context__


def test_invalid_list_element_type_in_subclass():
    @dataclass(frozen=True)
    class TstClass:
        a: int
        b: Optional[str]
        c: List[int]
        d: Dict[str, int]

    @dataclass(frozen=True)
    class TstClassMain:
        foo: TstClass

    with pytest.raises(FromDictTypeError) as e:
        opt = from_dict(TstClassMain, foo=dict(a=11, b=None, c=[1, 2, 3, "bad"], d={"a":1, "b": 2}), fd_check_types=True)

    assert str(e.value) == "For \"foo.c[3]\", expected <class 'int'> but found <class 'str'>"
    assert e.value.__suppress_context__


def test_invalid_dict_element_type_in_subclass():
    @dataclass(frozen=True)
    class TstClass:
        a: int
        b: Optional[str]
        c: List[int]
        d: Dict[str, int]

    @dataclass(frozen=True)
    class TstClassMain:
        foo: TstClass

    with pytest.raises(FromDictTypeError) as e:
        opt = from_dict(TstClassMain, foo=dict(a=11, b=None, c=[1, 2, 3, 4], d={"a":1, "b": 2, "C": "bad"}), fd_check_types=True)

    assert str(e.value) == "For \"foo.d['C']\", expected <class 'int'> but found <class 'str'>"
    assert e.value.__suppress_context__


def test_subscripted_attr_generics_work():
    @attr.s(auto_attribs=True)
    class KDict:
        a: int
        b: Optional[str]
        c: List[int]
        d: Dict[str, int]

    opt = from_dict(KDict, a=11, b=None, c=[1, 2, 3], d={"a":1, "b": 2})

    assert opt.a == 11
    assert opt.b is None
    assert opt.c == [1, 2, 3]
    assert opt.d == {"a":1, "b": 2}
    assert from_dict(KDict, a=11, b="hi", c=[1, 2, 3], d={"a":1, "b": 2}).b == "hi"


def test_dataclass_generics_work():
    @dataclass(frozen=True)
    class KDict:
        a: int
        b: Optional[str]
        c: List[int]
        d: Dict[str, int]

    opt = from_dict(KDict, a=11, b=None, c=[1, 2, 3], d={"a":1, "b": 2})

    assert opt.a == 11
    assert opt.b is None
    assert opt.c == [1, 2, 3]
    assert opt.d == {"a":1, "b": 2}

    opt = from_dict(KDict, dict(a=11, b="hi", c=[1, 2, 3], d={"a":1, "b": 2}))

    assert opt.a == 11
    assert opt.b == "hi"
    assert opt.c == [1, 2, 3]
    assert opt.d == {"a":1, "b": 2}


def test_list_of_structures_work(structures: Structures):
    @attr.s(auto_attribs=True)
    class KList:
        a: List[structures.inner_structure]

    val = {
        "a": [
            {"foo": 11, "bar": "Hi"},
            {"foo": 13, "bar": "Sup"},
            {"foo": -4, "bar": "Ya"},
        ]
    }
    structs = from_dict(KList, val, fd_check_types=not structures.does_type_validation)

    assert structs.a[0].foo == 11
    assert [el.foo for el in structs.a] == [11, 13, -4]
    assert [el.bar for el in structs.a] == ["Hi", "Sup", "Ya"]


def test_dict_with_substructure(structures: Structures):
    @attr.s(auto_attribs=True)
    class SubDict:
        a: Dict[int, structures.inner_structure]

    val = {
        "a": {
            11: {"foo": 11, "bar": "Hi"},
            13: {"foo": 13, "bar": "Sup"},
            -4: {"foo": -4, "bar": "Ya"},
        }
    }
    structs = from_dict(SubDict, val, fd_check_types=not structures.does_type_validation)
    assert all(k == v.foo for k, v in structs.a.items())


def test_union_works():
    @attr.s(auto_attribs=True)
    class UClass:
        a: Union[str, int]

    assert from_dict(UClass, {"a": "hello"}, fd_check_types=True).a == "hello"
    assert from_dict(UClass, {"a": 22}, fd_check_types=True).a == 22


def test_generic_dataclass():
    TField1 = TypeVar('TField1')
    TField2 = TypeVar('TField2')
    @dataclass(frozen=True)
    class Data1:
        value: int

    @dataclass(frozen=True)
    class Data2:
        value: str

    @dataclass(frozen=True)
    class TstClassMain(Generic[TField1, TField2]):
        field_1: TField1
        field_2: TField2
        field_3: str
        field_4: int

    v = from_dict(TstClassMain[Data1, Data2], field_1=dict(value=1), field_2={"value":"1"}, field_3="s", field_4=1, fd_check_types=True)
    assert isinstance(v.field_1, Data1)
    assert isinstance(v.field_2, Data2)
    assert isinstance(v.field_3, str)
    assert isinstance(v.field_4, int)

    v = from_dict(TstClassMain[Data2, Data1], field_1=dict(value="1"), field_2={"value":1}, field_3="s", field_4=1, fd_check_types=True)
    assert isinstance(v.field_1, Data2)
    assert isinstance(v.field_2, Data1)
    assert isinstance(v.field_3, str)
    assert isinstance(v.field_4, int)


def test_parent_generic_dataclass():
    TField1 = TypeVar('TField1')
    TField2 = TypeVar('TField2')
    @dataclass(frozen=True)
    class Data1:
        value: int

    @dataclass(frozen=True)
    class Data2:
        value: str

    @dataclass(frozen=True)
    class TstClassMainParent(Generic[TField1, TField2]):
        field_1: TField1
        field_2: TField2
        field_3: str
        field_4: int

    @dataclass(frozen=True)
    class TstClassMain(TstClassMainParent[Data1, Data2]):
        pass

    v = from_dict(TstClassMain, field_1=dict(value=1), field_2={"value":"1"}, field_3="s", field_4=1, fd_check_types=True)
    assert isinstance(v.field_1, Data1)
    assert isinstance(v.field_2, Data2)
    assert isinstance(v.field_3, str)
    assert isinstance(v.field_4, int)


def test_generic_norm_class():
    TParam1 = TypeVar('TParam1')
    TParam2 = TypeVar('TParam2')
    @dataclass(frozen=True)
    class Data1:
        value: int

    @dataclass(frozen=True)
    class Data2:
        value: str

    class TstClassMain(Generic[TParam1, TParam2]):
        def __init__(self, param_1: TParam1, param_2: TParam2, param_3: str, param_4: int) -> None:
            self.field_1 = param_1
            self.field_2 = param_2
            self.field_3 = param_3
            self.field_4 = param_4

    v = from_dict(TstClassMain[Data1, Data2], param_1={"value":1}, param_2=dict(value="1"), param_3="s", param_4=1, fd_check_types=True)
    assert isinstance(v.field_1, Data1)
    assert isinstance(v.field_2, Data2)
    assert isinstance(v.field_3, str)
    assert isinstance(v.field_4, int)

    v = from_dict(TstClassMain[Data2, Data1], param_1={"value":"1"}, param_2=dict(value=1), param_3="s", param_4=1, fd_check_types=True)
    assert isinstance(v.field_1, Data2)
    assert isinstance(v.field_2, Data1)
    assert isinstance(v.field_3, str)
    assert isinstance(v.field_4, int)


def test_parent_generic_norm_class():
    TParam1 = TypeVar('TParam1')
    TParam2 = TypeVar('TParam2')
    @dataclass(frozen=True)
    class Data1:
        value: int

    @dataclass(frozen=True)
    class Data2:
        value: str

    class TstClassMainParent(Generic[TParam1, TParam2]):
        def __init__(self, param_1: TParam1, param_2: TParam2, param_3: str, param_4: int) -> None:
            self.field_1 = param_1
            self.field_2 = param_2
            self.field_3 = param_3
            self.field_4 = param_4

    class TstClassMain(TstClassMainParent[Data1, Data2]):
        pass

    v = from_dict(TstClassMain, param_1={"value":1}, param_2=dict(value="1"), param_3="s", param_4=1, fd_check_types=True)
    assert isinstance(v.field_1, Data1)
    assert isinstance(v.field_2, Data2)
    assert isinstance(v.field_3, str)
    assert isinstance(v.field_4, int)


def test_generic_dataclass_with_generic_fields():
    TField1 = TypeVar('TField1')
    TField2 = TypeVar('TField2')
    @dataclass(frozen=True)
    class Data1:
        value: int

    @dataclass(frozen=True)
    class Data2:
        value: str

    @dataclass(frozen=True)
    class TstClassMain(Generic[TField1, TField2]):
        f_1: Optional[TField1]
        f_2: List[TField2]

    v = from_dict(TstClassMain[Data1, Data2], {"f_1": {"value":1}, "f_2": [{"value":"1"}] }, fd_check_types=True)
    assert isinstance(v.f_1, Data1)
    assert isinstance(v.f_2, list)
    assert isinstance(v.f_2[0], Data2)

    v = from_dict(TstClassMain[Data1, Data2], {"f_1": None, "f_2": [{"value":"1"}] }, fd_check_types=True)
    assert v.f_1 is None
    assert isinstance(v.f_2, list)
    assert isinstance(v.f_2[0], Data2)

    v = from_dict(TstClassMain[Data2, Data1], {"f_1": {"value":"1"}, "f_2": [{"value":1}]}, fd_check_types=True)
    assert isinstance(v.f_1, Data2)
    assert isinstance(v.f_2, list)
    assert isinstance(v.f_2[0], Data1)

    v = from_dict(TstClassMain['Data2', 'Data1'], {"f_1": {"value":"1"}, "f_2": [{"value":1}]}, fd_local_ns=locals(), fd_check_types=True)
    assert isinstance(v.f_1, Data2)
    assert isinstance(v.f_2, list)
    assert isinstance(v.f_2[0], Data1)


def test_parent_generic_dataclass_with_generic_fields():
    TField1 = TypeVar('TField1')
    TField2 = TypeVar('TField2')
    @dataclass(frozen=True)
    class Data1:
        value: int

    @dataclass(frozen=True)
    class Data2:
        value: str

    @dataclass(frozen=True)
    class TstClassMainParent(Generic[TField1, TField2]):
        f_1: Optional[TField1]
        f_2: List[TField2]

    class TstClassMain(TstClassMainParent[Data1, Data2]):
        pass

    v = from_dict(TstClassMain, {"f_1": {"value":1}, "f_2": [{"value":"1"}] }, fd_check_types=True)
    assert isinstance(v.f_1, Data1)
    assert isinstance(v.f_2, list)
    assert isinstance(v.f_2[0], Data2)

    v = from_dict(TstClassMain, {"f_1": None, "f_2": [{"value":"1"}] }, fd_check_types=True)
    assert v.f_1 is None
    assert isinstance(v.f_2, list)
    assert isinstance(v.f_2[0], Data2)

def test_error_on_unknown_args():
    # Verify these don't raise an error
    from_dict(SubTestDictDataclass, foo=1, bar="s", fd_check_types=True)
    from_dict(SubTestDictDataclass, foo=1, bar="s", fd_check_types=True, fd_error_on_unknown=True, fd_copy_unknown=False)
    from_dict(SubTestDictDataclass, foo=1, bar="s", sam=1, fd_check_types=True)

    with pytest.raises(FromDictUnknownArgsError) as e:
        from_dict(SubTestDictDataclass, foo=1, bar="s", sam=1, fd_error_on_unknown=True, fd_copy_unknown=False, fd_check_types=True)
    
    # Verify these don't raise an error
    from_dict(MainTestDictDataclass, foo=1, baz=dict(foo=1, bar="s"), fd_check_types=True)
    from_dict(MainTestDictDataclass, foo=1, baz=dict(foo=1, bar="s"), fd_check_types=True, fd_error_on_unknown=True, fd_copy_unknown=False)
    from_dict(MainTestDictDataclass, foo=1, baz=dict(foo=5, bar="s", sam=2), fd_check_types=True)

    with pytest.raises(FromDictUnknownArgsError) as e:
        from_dict(MainTestDictDataclass, foo=1, baz=dict(foo=5, bar="s", sam=2), fd_error_on_unknown=True, fd_copy_unknown=False, fd_check_types=True)
    
    @dataclass
    class ClassWithDefaults:
        foo: int = 1
        bar: str = "default"
          
    # Verify these don't raise an error
    from_dict(ClassWithDefaults, foo=3, fd_check_types=True)
    from_dict(ClassWithDefaults, fd_check_types=True, fd_error_on_unknown=True, fd_copy_unknown=False)
    from_dict(ClassWithDefaults, sam=1, fd_check_types=True)

    with pytest.raises(FromDictUnknownArgsError) as e:
        from_dict(ClassWithDefaults, sam=1, fd_error_on_unknown=True, fd_copy_unknown=False, fd_check_types=True)


import copy
import datetime
from contextlib import contextmanager
from dataclasses import dataclass
from typing import Any, Dict, List, Optional, Union, Type, Callable, TypeVar, Generic, Literal

from from_dict import FromDictTypeError, from_dict

import attr
import pytest

# . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .
# Classes used by the tests

class NormalClass:
    def __init__(self, v: str) -> None:
        self.value = v

    def __eq__(self, other: object) -> bool:
        return isinstance(other, NormalClass) and self.value == other.value


@dataclass(frozen=True)
class DataClass:
    field: str


@attr.s(auto_attribs=True)
class AttrClass:
    attrib: str

TTestType = TypeVar("TTestType")
TSelfRef = TypeVar("TSelfRef")

@dataclass
class ClassBase(Generic[TTestType, TSelfRef]):
    normal: TTestType
    optional: Optional[TTestType]
    union: Union[str, NormalClass, DataClass, TTestType]
    any: Any
    self_ref: Optional[TSelfRef]

    list_normal: List[TTestType]
    list_optional: List[Optional[TTestType]]
    list_union: List[Union[str, NormalClass, DataClass, TTestType]]
    list_any: List[Any]
    list_self_ref: List[TSelfRef]
    
    dict_normal: Dict[str, TTestType]
    dict_optional: Dict[str, Optional[TTestType]]
    dict_union: Dict[str, Union[str, NormalClass, DataClass, TTestType]]
    dict_any: Dict[str, Any]
    dict_self_ref: Dict[str, TSelfRef]


ClassPrimitives = ClassBase[int, "ClassPrimitives"]
ClassDict = ClassBase[Dict[str,str], "ClassDict"]
ClassDictSimple = ClassBase[dict, "ClassDictSimple"]
ClassList = ClassBase[List[str], "ClassList"]
ClassDataClass = ClassBase[DataClass, "ClassDataClass"]
ClassNormalClass = ClassBase[NormalClass, "ClassNormalClass"]
ClassAttrClass = ClassBase[AttrClass, "ClassAttrClass"]
ClassListDataClass = ClassBase[List[DataClass], "ClassListDataClass"]
ClassDictDataClass = ClassBase[Dict[str, DataClass], "ClassDictDataClass"]
ClassLiteral = ClassBase[Literal["my-literal"], "ClassLiteral"]
ClassMultiLiteral = ClassBase[Literal[2, 3, 5, 7, 11], "ClassMultiLiteral"]

# . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .
# helper functions

def initialize(cls, init_func, any_init = None):
        any_init = any_init or init_func
        return cls(
        normal = init_func(),
        optional = init_func(),
        union = init_func(),
        any = any_init(), # If this is a dict, it will not be converted to a dataclass
        self_ref = None,

        list_normal = [init_func()],
        list_optional = [None, init_func()],
        list_union = ["first", init_func()],
        list_any = [datetime.datetime.now().date(), {"v": "VALUE"}, {"field": "ANY"}],
        list_self_ref = [],

        dict_normal = {"k": init_func()},
        dict_optional = {"K1": None, "K2":init_func()},
        dict_union = {"k1": "first", "k2":init_func()},
        dict_any = {"k1":datetime.datetime.now().date(), "k2":{"v": "VALUE"}, "k3":{"field": "ANY"}},
        dict_self_ref = {},
    )


@contextmanager
def expect_type_error():
    try:
        yield None
    except FromDictTypeError:
        return
    raise AssertionError("Expected 'FromDictTypeError' to be raised")


# . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .
# PyTest framework
@dataclass
class NormTestParams:
    cls: Type
    dic_init_func: Callable
    expected_init_func: Callable


@pytest.fixture(params=[
    pytest.param(NormTestParams(ClassPrimitives, lambda: 4, lambda: 4), id="primitives"),
    pytest.param(NormTestParams(ClassDict, lambda: {"d":"Vvv"}, lambda: {"d":"Vvv"}), id="dict"),
    pytest.param(NormTestParams(ClassDictSimple, lambda: {"d":"Vvv"}, lambda: {"d":"Vvv"}), id="dict-2"),
    pytest.param(NormTestParams(ClassList, lambda: ["1", "2"], lambda: ["1", "2"]), id="list"),
    
    pytest.param(NormTestParams(ClassDataClass, lambda: {"field": "value"}, lambda: DataClass("value")), id="data-class"),
    pytest.param(NormTestParams(ClassNormalClass, lambda: {"v": "value"}, lambda: NormalClass("value")), id="norm-class"),
    pytest.param(NormTestParams(ClassAttrClass, lambda: {"attrib": "value"}, lambda: AttrClass("value")), id="attr-class"),

    pytest.param(NormTestParams(ClassListDataClass, lambda: [{"field": "value"}], lambda: [DataClass("value")]), id="list-data-class"),
    pytest.param(NormTestParams(ClassDictDataClass, lambda: {"KEY":{"field": "value"}}, lambda: {"KEY":DataClass("value")}), id="dict-data-class"),

    pytest.param(NormTestParams(ClassLiteral, lambda: "my-literal", lambda: "my-literal"), id="literal-single"),
    pytest.param(NormTestParams(ClassMultiLiteral, lambda: 3, lambda: 3), id="literal-multi-1"),
    pytest.param(NormTestParams(ClassMultiLiteral, lambda: 11, lambda: 11), id="literal-multi-2"),
])
def norm_params(request):
    yield request.param

@dataclass
class NegativeTestParams:
    cls: Type
    dic_init_func: Callable
    bad_init_func: Callable


@pytest.fixture(params=[
    pytest.param(NegativeTestParams(ClassPrimitives, lambda: 4, datetime.datetime.now), id="primitives"),
    pytest.param(NegativeTestParams(ClassDict, lambda: {"d":"Vvv"}, datetime.datetime.now), id="dict"),
    pytest.param(NegativeTestParams(ClassDictSimple, lambda: {"d":"Vvv"}, datetime.datetime.now), id="dict-2"),
    pytest.param(NegativeTestParams(ClassList, lambda: ["1", "2"], datetime.datetime.now), id="list"),

    pytest.param(NegativeTestParams(ClassDataClass, lambda: {"field": "value"}, datetime.datetime.now), id="data-class"),
    pytest.param(NegativeTestParams(ClassNormalClass, lambda: {"v": "value"}, datetime.datetime.now), id="norm-class"),
    pytest.param(NegativeTestParams(ClassAttrClass, lambda: {"attrib": "value"}, datetime.datetime.now), id="attr-class"),
    
    pytest.param(NegativeTestParams(ClassListDataClass, lambda: [{"field": "value"}], datetime.datetime.now), id="list-data-class"),
    pytest.param(NegativeTestParams(ClassDictDataClass, lambda: {"KEY":{"field": "value"}}, datetime.datetime.now), id="dict-data-class"),

    pytest.param(NegativeTestParams(ClassLiteral, lambda: "my-literal", lambda: 1), id="literal-single"),
    pytest.param(NegativeTestParams(ClassMultiLiteral, lambda: 3, lambda: 1), id="literal-multi-1"),
    pytest.param(NegativeTestParams(ClassMultiLiteral, lambda: 5, lambda: 10), id="literal-multi-2"),
])
def negative_params(request):
    yield request.param

# . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .
# The tests

def test_basic(norm_params: NormTestParams):
    data_1 = initialize(dict, norm_params.dic_init_func)
    obj_1 = from_dict(norm_params.cls, data_1, fd_check_types=True)
    
    expected_1 = initialize(norm_params.cls, norm_params.expected_init_func, norm_params.dic_init_func)
    assert obj_1 == expected_1


def test_self_ref(norm_params: NormTestParams):
    data_1 = initialize(dict, norm_params.dic_init_func)
    obj_1 = from_dict(norm_params.cls, data_1, fd_check_types=True)
    
    expected_1 = initialize(norm_params.cls, norm_params.expected_init_func, norm_params.dic_init_func)
    assert obj_1 == expected_1
    
    # . . . . . . . . . . . . . . . . . . . . . . . . . . .
    data_2 = copy.deepcopy(data_1)
    expected_2 = copy.deepcopy(expected_1)
    data_2['self_ref'] = data_1
    expected_2.self_ref = expected_1

    obj_2 = from_dict(norm_params.cls, data_2, fd_check_types=True)
    assert obj_2 == expected_2
    
    # . . . . . . . . . . . . . . . . . . . . . . . . . . .
    data_3 = copy.deepcopy(data_2)
    expected_3 = copy.deepcopy(expected_2)
    data_3['list_self_ref'] = [data_1]
    expected_3.list_self_ref = [expected_1]

    obj_3 = from_dict(norm_params.cls, data_3, fd_check_types=True)
    assert obj_3 == expected_3
    
    # . . . . . . . . . . . . . . . . . . . . . . . . . . .
    data_4 = copy.deepcopy(data_3)
    expected_4 = copy.deepcopy(expected_3)
    data_4['dict_self_ref'] = {"k1":data_1, "k2":data_2}
    expected_4.dict_self_ref = {"k1":expected_1, "k2":expected_2}

    obj_4 = from_dict(norm_params.cls, data_4, fd_check_types=True)
    assert obj_4 == expected_4


def test_union(norm_params: NormTestParams):
    data: dict = initialize(dict, norm_params.dic_init_func)
    obj = from_dict(norm_params.cls, data, fd_check_types=True)
    
    expected = initialize(norm_params.cls, norm_params.expected_init_func, norm_params.dic_init_func)
    assert obj == expected
    
    # . . . . . . . . . . . . . . . . . . . . . . . . . . .
    data['union'] = "str"
    expected.union = "str"

    obj = from_dict(norm_params.cls, data, fd_check_types=True)
    assert obj == expected
    
    # . . . . . . . . . . . . . . . . . . . . . . . . . . .
    data['union'] = NormalClass("val")
    expected.union = NormalClass("val")

    obj = from_dict(norm_params.cls, data, fd_check_types=True)
    assert obj == expected
    
    # . . . . . . . . . . . . . . . . . . . . . . . . . . .
    data['union'] = {"v": "val"}
    expected.union = NormalClass("val")
    
    obj = from_dict(norm_params.cls, data, fd_check_types=True)
    assert obj == expected
    
    # . . . . . . . . . . . . . . . . . . . . . . . . . . .
    data['union'] = DataClass("FIELD")
    expected.union = DataClass("FIELD")

    obj = from_dict(norm_params.cls, data, fd_check_types=True)
    assert obj == expected
    
    # . . . . . . . . . . . . . . . . . . . . . . . . . . .
    data['union'] = {"field": "FIELD"}
    expected.union = DataClass("FIELD")
    
    obj = from_dict(norm_params.cls, data, fd_check_types=True)
    assert obj == expected


def test_list_union(norm_params: NormTestParams):
    data: dict = initialize(dict, norm_params.dic_init_func)
    obj = from_dict(norm_params.cls, data, fd_check_types=True)
    
    expected = initialize(norm_params.cls, norm_params.expected_init_func, norm_params.dic_init_func)
    assert obj == expected
    
    # . . . . . . . . . . . . . . . . . . . . . . . . . . .
    data['list_union'].append("str")
    expected.list_union.append("str")

    obj = from_dict(norm_params.cls, data, fd_check_types=True)
    assert obj == expected
    
    # . . . . . . . . . . . . . . . . . . . . . . . . . . .
    data['list_union'].append(NormalClass("val"))
    expected.list_union.append(NormalClass("val"))

    obj = from_dict(norm_params.cls, data, fd_check_types=True)
    assert obj == expected
    
    # . . . . . . . . . . . . . . . . . . . . . . . . . . .
    data['list_union'].append({"v": "val"})
    expected.list_union.append(NormalClass("val"))
    
    obj = from_dict(norm_params.cls, data, fd_check_types=True)
    assert obj == expected
    
    # . . . . . . . . . . . . . . . . . . . . . . . . . . .
    data['list_union'].append(DataClass("FIELD"))
    expected.list_union.append(DataClass("FIELD"))

    obj = from_dict(norm_params.cls, data, fd_check_types=True)
    assert obj == expected
    
    # . . . . . . . . . . . . . . . . . . . . . . . . . . .
    data['list_union'].append({"field": "FIELD"})
    expected.list_union.append(DataClass("FIELD"))
    
    obj = from_dict(norm_params.cls, data, fd_check_types=True)
    assert obj == expected
    
    # . . . . . . . . . . . . . . . . . . . . . . . . . . .
    data['list_union'] = [NormalClass("val")]
    expected.list_union = [NormalClass("val")]

    obj = from_dict(norm_params.cls, data, fd_check_types=True)
    assert obj == expected
    
    # . . . . . . . . . . . . . . . . . . . . . . . . . . .
    data['list_union'] = [{"v": "val"}]
    expected.list_union = [NormalClass("val")]
    
    obj = from_dict(norm_params.cls, data, fd_check_types=True)
    assert obj == expected
    
    # . . . . . . . . . . . . . . . . . . . . . . . . . . .
    data['list_union'] = [DataClass("FIELD")]
    expected.list_union = [DataClass("FIELD")]

    obj = from_dict(norm_params.cls, data, fd_check_types=True)
    assert obj == expected
    
    # . . . . . . . . . . . . . . . . . . . . . . . . . . .
    data['list_union'] = [{"field": "FIELD"}]
    expected.list_union = [DataClass("FIELD")]
    
    obj = from_dict(norm_params.cls, data, fd_check_types=True)
    assert obj == expected


def verify_negative_basic(negative_params: NegativeTestParams):
    data = initialize(dict, negative_params.dic_init_func)
    data["normal"] = negative_params.bad_init_func()

    with expect_type_error():
        from_dict(negative_params.cls, data, fd_check_types=True)
    
    # . . . . . . . . . . . . . . . . . . . . . . . . . . .
    data = initialize(dict, negative_params.dic_init_func)
    data["optional"] = negative_params.bad_init_func()

    with expect_type_error():
        from_dict(negative_params.cls, data, fd_check_types=True)
    
    # . . . . . . . . . . . . . . . . . . . . . . . . . . .
    data = initialize(dict, negative_params.dic_init_func)
    data["union"] = negative_params.bad_init_func()

    with expect_type_error():
        from_dict(negative_params.cls, data, fd_check_types=True)
    
    # . . . . . . . . . . . . . . . . . . . . . . . . . . .
    data = initialize(dict, negative_params.dic_init_func)
    data["self_ref"] = negative_params.bad_init_func()

    with expect_type_error():
        from_dict(negative_params.cls, data, fd_check_types=True)
    

def test_negative_list(negative_params: NegativeTestParams):
    data = initialize(dict, negative_params.dic_init_func)
    data["list_normal"] = [negative_params.bad_init_func()]

    with expect_type_error():
        from_dict(negative_params.cls, data, fd_check_types=True)
    
    # . . . . . . . . . . . . . . . . . . . . . . . . . . .
    data = initialize(dict, negative_params.dic_init_func)
    data["list_optional"] = [negative_params.bad_init_func()]

    with expect_type_error():
        from_dict(negative_params.cls, data, fd_check_types=True)
    
    # . . . . . . . . . . . . . . . . . . . . . . . . . . .
    data = initialize(dict, negative_params.dic_init_func)
    data["list_union"] = [negative_params.bad_init_func()]

    with expect_type_error():
        from_dict(negative_params.cls, data, fd_check_types=True)
    
    # . . . . . . . . . . . . . . . . . . . . . . . . . . .
    data = initialize(dict, negative_params.dic_init_func)
    data["list_self_ref"] = [negative_params.bad_init_func()]

    with expect_type_error():
        from_dict(negative_params.cls, data, fd_check_types=True)


def test_negative_dict(negative_params: NegativeTestParams):
    data: dict = initialize(dict, negative_params.dic_init_func)
    data["dict_normal"] = {"k":negative_params.bad_init_func()}

    with expect_type_error():
        from_dict(negative_params.cls, data, fd_check_types=True)
    
    # . . . . . . . . . . . . . . . . . . . . . . . . . . .
    data = initialize(dict, negative_params.dic_init_func)
    data["dict_optional"] = {"k":negative_params.bad_init_func()}

    with expect_type_error():
        from_dict(negative_params.cls, data, fd_check_types=True)
    
    # . . . . . . . . . . . . . . . . . . . . . . . . . . .
    data = initialize(dict, negative_params.dic_init_func)
    data["dict_union"] = {"k":negative_params.bad_init_func()}

    with expect_type_error():
        from_dict(negative_params.cls, data, fd_check_types=True)
    
    # . . . . . . . . . . . . . . . . . . . . . . . . . . .
    data = initialize(dict, negative_params.dic_init_func)
    data["dict_self_ref"] = {"k":negative_params.bad_init_func()}

    with expect_type_error():
        from_dict(negative_params.cls, data, fd_check_types=True)

    # . . . . . . . . . . . . . . . . . . . . . . . . . . .
    data = initialize(dict, negative_params.dic_init_func)
    data["dict_normal"] = {123:negative_params.dic_init_func()}

    with expect_type_error():
        from_dict(negative_params.cls, data, fd_check_types=True)
    
    # . . . . . . . . . . . . . . . . . . . . . . . . . . .
    data = initialize(dict, negative_params.dic_init_func)
    data["dict_optional"] = {123:negative_params.dic_init_func()}

    with expect_type_error():
        from_dict(negative_params.cls, data, fd_check_types=True)
    
    # . . . . . . . . . . . . . . . . . . . . . . . . . . .
    data = initialize(dict, negative_params.dic_init_func)
    data["dict_union"] = {123:negative_params.dic_init_func()}

    with expect_type_error():
        from_dict(negative_params.cls, data, fd_check_types=True)
    
    # . . . . . . . . . . . . . . . . . . . . . . . . . . .
    data = initialize(dict, negative_params.dic_init_func)
    data["dict_self_ref"] = {123:data.copy()}

    with expect_type_error():
        from_dict(negative_params.cls, data, fd_check_types=True)
    
    # . . . . . . . . . . . . . . . . . . . . . . . . . . .
    data = initialize(dict, negative_params.dic_init_func)
    data["dict_any"] = {123:negative_params.dic_init_func()}

    with expect_type_error():
        from_dict(negative_params.cls, data, fd_check_types=True)




import sys
from dataclasses import dataclass
from typing import Dict, List, Optional

from from_dict import from_dict

if sys.version_info[:2] >= (3, 9):
    LIST = list
else:
    LIST = List


@dataclass
class LinkListNode:
    name: str
    next: 'LinkListNode'

@dataclass
class TreeNode:
    name: str
    children: LIST['TreeNode']

@dataclass
class LinkListNode2:
    name: str
    next: Optional['LinkListNode2'] 


@dataclass
class DictNode:
    name: str
    children: Dict[str, 'DictNode']

def test_self_ref():
    data = {"name": "n1", "next": {"name": "n2", "next": None}}
    # Can't type check because next is not optional 
    node = from_dict(LinkListNode, data)
    assert node.name == "n1"
    assert node.next.name == "n2"


def test_self_ref_in_list():
    data = {"name": "n1", "children": [{"name": "n2", "children": []}]}
    node = from_dict(TreeNode, data, fd_check_types=True)
    assert node.name == "n1"
    assert node.children[0].name == "n2"


def test_self_ref_in_optional():
    data = {"name": "n1", "next": {"name": "n2", "next": None}}
    node = from_dict(LinkListNode2, data, fd_check_types=True)
    assert node.name == "n1"
    assert isinstance(node.next, LinkListNode2)
    assert node.next.name == "n2"


def test_self_ref_in_dict():
    data = {
        "name": "n1", 
        "children": {
            "id-123": {"name": "n2", "children": {}},
            "id-456": {"name": "n3", "children": {}}
        }
    }
    node = from_dict(DictNode, data, fd_check_types=True)
    assert node.name == "n1"
    assert node.children["id-123"].name == "n2"
    assert node.children["id-456"].name == "n3"




import sys
from dataclasses import dataclass
from typing import Dict, List, Optional

from from_dict import from_dict

if sys.version_info[:2] >= (3, 9):
    LIST = list
else:
    LIST = List


def test_local_self_ref():
    @dataclass
    class Node:
        name: str
        next: 'Node'

    data = {"name": "n1", "next": {"name": "n2", "next": None}}
    node = from_dict(Node, data, fd_local_ns=locals())
    assert node.name == "n1"
    assert node.next.name == "n2"

def test_local_self_ref_in_list():
    @dataclass
    class Node:
        name: str
        children: LIST['Node']

    data = {"name": "n1", "children": [{"name": "n2", "children": []}]}
    node = from_dict(Node, data, fd_check_types=True, fd_local_ns=locals())
    assert node.name == "n1"
    assert node.children[0].name == "n2"


def test_local_self_ref_in_optional():
    @dataclass
    class Node:
        name: str
        next: Optional['Node']

    data = {"name": "n1", "next": {"name": "n2", "next": None}}
    node = from_dict(Node, data, fd_check_types=True, fd_local_ns=locals())
    assert node.name == "n1"
    assert isinstance(node.next, Node)
    assert node.next.name == "n2"

    
def test_local_self_ref_in_dict():
    @dataclass
    class Node:
        name: str
        children: Dict[str, 'Node']

    data = {
        "name": "n1", 
        "children": {
            "id-123": {"name": "n2", "children": {}},
            "id-456": {"name": "n3", "children": {}}
        }
    }
    node = from_dict(Node, data, fd_check_types=True, fd_local_ns=locals())
    assert node.name == "n1"
    assert node.children["id-123"].name == "n2"
    assert node.children["id-456"].name == "n3"
